from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum

from bson import ObjectId


class DecisionMakerDepartment(Enum):
    """The department of the decision maker."""

    it_RND = "IT / R&D"
    sales_marketing_communication = "Sales / Marketing / Communication"
    legal = "Legal"
    finance = "Finance"
    ceo_board = "CEO / Board"
    purchase_logistic_safety = "Purchase / Logistic / Safety"
    human_resources = "Human Resources"
    customer_support = "Customer Support"
    enviromental_esg = "Environmental / ESG"
    administration = "Administration"
    project_management = "Project Management"
    not_working = "Not working anymore / Not a person / Not a role"
    other = "Other"

    @classmethod
    def get_all_values(cls):
        """Return all possible values for this enum as a list of enum."""
        return [v for v in cls.__dict__.values() if isinstance(v, cls)]

    @classmethod
    def get_all_values_str(cls) -> list[str]:
        """Return all possible values for this enum as a list of str."""
        return [v.value for v in cls.__dict__.values() if isinstance(v, cls)]


responsibilities_level_mapper: dict[int, str] = {
    1: "C-Level / Director / Board / Chief / Owner / Founder",
    2: "Manager / Middle management",
    3: "Employee (Senior)",
    4: "Employee (Junior/Medior)",
    5: "Other",
    6: "Not working anymore / Not a person / Not a role",
}


@dataclass(slots=True)
class DecisionMaker:
    # Foreign key to the `company` collection
    company_id: str
    firstname: str
    lastname: str
    source: str
    # The raw string from the source, will cleaned and classified into `function_string`
    raw_function_string: str
    last_seen: datetime
    last_email_check: datetime | None = None
    _id: ObjectId = field(default_factory=ObjectId)
    function_string: str | None = None
    email: str | None = None
    email_score: int | None = None
    language: str | None = None
    linkedin_url: str | None = None
    department: str | None = None
    responsibility_level_code: int | None = None
    responsibility_level_formatted: str | None = None



import os
import time
from datetime import datetime
from typing import cast

import logfire
import requests
from dotenv import load_dotenv
from typesense.client import Client as TypesenseClient
from typesense.collection import Collection

load_dotenv()

EMBEDDING_MODEL = "text-embedding-3-small"
EMBEDDING_MODEL_EXPENSIVE = "text-embedding-3-large"
EMBEDDING_CONTEXT_MAX_LENGTH = 8191
EMBEDDING_ENCODING = "cl100k_base"
# For the model name to typesense's namming convention
EMBEDDING_MODEL_TYPESENSE_NAME = f"openai/{EMBEDDING_MODEL}"
EXPENSIVE_EMBEDDING_MODEL_TYPESENSE_NAME = f"openai/{EMBEDDING_MODEL_EXPENSIVE}"
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")


class TypesenseManager:
    """
    This class is a client to interact with the Typesense search engine server (deployed in InfraV2).
    You can then is the client outside of the class to interact with the Typesense server.

    It also take care of creating the collections if they don't exist.
    ! Please maintain the schemas in the same file as the TypesenseManager class. !

    :attribute typesense_url: The typesense server URL.
    :attribute client: The typesense client object to interact with the typesense server.
    :attribute collection_nace_codes: The nace_codes collection object.
    :attribute collection_companies_names: The companies_names collection object.
    :attribute collection_descriptions: The descriptions collection object. (embedded desc for all websites & NACEs)
    """

    def __init__(
        self,
        host: str = os.getenv("TYPESENSE_HOST", ""),
        port: int = int(os.getenv("TYPESENSE_PORT", 443)),  # Default port is 443 for HTTPS
        api_key: str = os.getenv("TYPESENSE_API_KEY", ""),
        protocol: str = "https",
        pretty_logging: bool = False,
    ):
        """
        :param host: Typesense's server host
        :param port: Typesense's server port
        :param api_key: Typesense's API key
        :param protocol: Typesense's server protocol (http or https)
        """
        self.typesense_url = f"https://{host}"
        self.client: TypesenseClient = TypesenseClient(
            {
                "api_key": api_key,
                "nodes": [{"host": host, "port": port, "protocol": protocol}],
                "connection_timeout_seconds": 60 * 20,  # 20 minutes max
            }
        )

        # 1. Check the API's health
        self._check_server_heatlh()
        # 2. Create collections if they don't exist
        self._create_collections()

        # Expose collections
        self.collection_nace_codes: Collection = cast(Collection, self.client.collections["nace_codes"])
        self.collection_companies_names: Collection = cast(Collection, self.client.collections["companies_names"])
        self.collection_descriptions: Collection = cast(Collection, self.client.collections["descriptions"])
        logfire.debug("Typesense Manager ready!")

    def _check_server_heatlh(self) -> None:
        """Check if typesense server is up and running."""
        response = requests.get(f"{self.typesense_url}/health")
        response.raise_for_status()
        if not response.json()["ok"]:
            raise Exception("Typesense is not running!")
        logfire.info("Typesense server is up and running!")

    def _create_collections(self) -> None:
        """Create a Typesense schema so MongoDB's collection can be interpreted."""
        collections = self.client.collections.retrieve()
        existing_collections_name = set([col["name"] for col in collections])
        for collection in [naces_collection_schema, companies_names_collection_schema, descriptions_schema]:
            if collection["name"] not in existing_collections_name:
                self.client.collections.create(collection)  # type: ignore
                logfire.warning(f"Typesense collection {collection['name']} created!")
                continue
            logfire.debug(f"Typesense collection {collection['name']} already exists!")


def datetime_to_unix_timestamp(date: datetime) -> int:
    """
    Convert datetime to unix timestamp as int.

    This is the format that Typesense is expecting for datetime data.
    See: https://typesense.org/docs/guide/tips-for-searching-common-types-of-data.html#searching-for-null-or-empty-values
    """
    return int(time.mktime(date.timetuple()))


def unix_timestamp_to_datetime(timestamp: int) -> datetime:
    return datetime.fromtimestamp(timestamp)


# Define the schemas for each collection

# Contains description for all companies with a website or a NACE code
descriptions_schema = {
    "name": "descriptions",
    "enable_nested_fields": True,
    "fields": [
        # ---- Imported data ----
        {"optional": False, "name": "id", "type": "string"},
        {"optional": False, "name": "description_short", "type": "string"},
        {"optional": False, "name": "description_long", "type": "string"},
        {"optional": False, "name": "source", "type": "string"},
        {"optional": False, "name": "source_value", "type": "string"},
        # ---- Embeddings ----
        {
            "optional": True,
            "name": "description_short_embedding",
            "type": "float[]",
            "embed": {
                "from": ["description_short"],
                "model_config": {"model_name": EXPENSIVE_EMBEDDING_MODEL_TYPESENSE_NAME, "api_key": OPENAI_API_KEY},
            },
        },
        {
            "optional": True,
            "name": "description_long_embedding",
            "type": "float[]",
            "embed": {
                "from": ["description_long"],
                "model_config": {"model_name": EXPENSIVE_EMBEDDING_MODEL_TYPESENSE_NAME, "api_key": OPENAI_API_KEY},
            },
        },
    ],
}

# Contains all the naces codes, their labels and translations
naces_collection_schema = {
    "name": "nace_codes",
    "enable_nested_fields": True,
    "fields": [
        # ---- Imported data ----
        {"optional": False, "name": "country", "type": "string"},
        {"optional": False, "name": "level", "type": "int32"},
        {"optional": False, "name": "code", "type": "string"},
        {"optional": False, "name": "section_label", "type": "string"},
        {"optional": False, "name": "section_code", "type": "string"},
        {"optional": False, "name": "label_en", "type": "string"},
        {"optional": False, "name": "label_en_extended", "type": "string"},
        {"optional": True, "name": "label_fr", "type": "string"},
        {"optional": True, "name": "label_fr_extended", "type": "string"},
        {"optional": True, "name": "label_nl", "type": "string"},
        {"optional": True, "name": "label_nl_extended", "type": "string"},
        # ---- Embeddings ----
        {
            "optional": True,
            "name": "label_en_embedding",
            "type": "float[]",
            "embed": {
                "from": ["label_en"],
                "model_config": {"model_name": EMBEDDING_MODEL_TYPESENSE_NAME, "api_key": OPENAI_API_KEY},
            },
        },
        {
            "optional": True,
            "name": "label_en_extended_embedding",
            "type": "float[]",
            "embed": {
                "from": ["label_en_extended"],
                "model_config": {"model_name": EMBEDDING_MODEL_TYPESENSE_NAME, "api_key": OPENAI_API_KEY},
            },
        },
        {
            "optional": True,
            "name": "label_fr_embedding",
            "type": "float[]",
            "embed": {
                "from": ["label_fr"],
                "model_config": {"model_name": EMBEDDING_MODEL_TYPESENSE_NAME, "api_key": OPENAI_API_KEY},
            },
        },
        {
            "optional": True,
            "name": "label_fr_extended_embedding",
            "type": "float[]",
            "embed": {
                "from": ["label_fr_extended"],
                "model_config": {"model_name": EMBEDDING_MODEL_TYPESENSE_NAME, "api_key": OPENAI_API_KEY},
            },
        },
        {
            "optional": True,
            "name": "label_nl_embedding",
            "type": "float[]",
            "embed": {
                "from": ["label_nl"],
                "model_config": {"model_name": EMBEDDING_MODEL_TYPESENSE_NAME, "api_key": OPENAI_API_KEY},
            },
        },
        {
            "optional": True,
            "name": "label_nl_extended_embedding",
            "type": "float[]",
            "embed": {
                "from": ["label_nl_extended"],
                "model_config": {"model_name": EMBEDDING_MODEL_TYPESENSE_NAME, "api_key": OPENAI_API_KEY},
            },
        },
    ],
}

# Contains all the companies and establishment names for name matching puprose
companies_names_collection_schema = {
    "name": "companies_names",
    "enable_nested_fields": True,
    "fields": [
        {"optional": False, "name": "_id", "type": "string"},
        {"optional": True, "name": "name", "type": "string"},
        {"optional": True, "name": "name_fr", "type": "string"},
        {"optional": True, "name": "name_nl", "type": "string"},
        {"optional": True, "name": "name_de", "type": "string"},
        {"optional": True, "name": "website", "type": "string"},
        {"optional": True, "name": "establishments.name", "type": "string[]"},
        {"optional": True, "name": "establishments.name_fr", "type": "string[]"},
        {"optional": True, "name": "establishments.name_nl", "type": "string[]"},
        {"optional": True, "name": "establishments.name_de", "type": "string[]"},
    ],
}


if __name__ == "__main__":
    typesense_manager = TypesenseManager()
    print("OK!")



from deepparse.parser import AddressParser
from deepparse.parser.formatted_parsed_address import FormattedParsedAddress

from inoopa_utils.custom_types.exceptions import AddressParserError
from inoopa_utils.custom_types.companies import ParsedAddress, Address

# Make it a global variable to avoid loading the model each time the function is called
PARSER = AddressParser(model_type="bpemb")


def parse_address(address: Address) -> Address:
    """Parse an Address object, update it with the parsed info and return the updated Address."""
    if address.string_address is None:
        return address
    parsed_address = parse_string_address(address.string_address)

    address.street = parsed_address.street
    address.number = parsed_address.number
    address.city = parsed_address.city
    address.postal_code = parsed_address.postal_code

    return address


def parse_string_address(address: str) -> ParsedAddress:
    """
    Parse a string address and return a ParsedAddress object.

    :param address: The address to parse in string format.
    :return: A ParsedAddress object.

    :raises AddressParserError: If the address parser fails to parse the address.
    """
    parsed_address = PARSER(address)
    if not isinstance(parsed_address, FormattedParsedAddress):
        raise AddressParserError(f"The address parser failed to parse the address: {address}")
    # Store the parsed address in a dataclass because the default parsed address type is not well structured
    parsed_address = ParsedAddress(
        street=parsed_address.StreetName,
        number=parsed_address.StreetNumber,
        city=parsed_address.Municipality,
        postal_code=parsed_address.PostalCode,
    )
    return parsed_address


if __name__ == "__main__":
    address = parse_string_address("1 rue de la paix 75001 Paris")
    print(address)



"""
This file contains common filters we use in our MongoDB question.
As MongoDB filters are dicts, you can unpack them like this:

    WITH_WEBSITE = {"best_website": {"$ne": None}}
    WITH_PHONE = {"best_phone": {"$ne": None}}

    results = collection.find({ **WITH_WEBSITE, **WITH_PHONE })
"""

from inoopa_utils.custom_types.addresses import Country, RegionBe
from inoopa_utils.custom_types.legal_filters import LegalFormCategory

# --- General ---
ACTIVE_COMPANIES = {"status": "Active", "legal_situation": "Normal situation"}

# --- Contact info ---
WITH_WEBSITE = {"best_website": {"$ne": None}}
WITH_PHONE = {"best_phone": {"$ne": None}}
WITH_PHONE_NO_DNCM = {"best_phone": {"$ne": None}, "best_phone.phone": {"$ne": "DO_NOT_CALL_ME"}}

# --- Activities sector (NACEs) ---
NO_LAWYER_NO_RELGIOUS_ORGS = {
    "$nor": [
        # Lawyers
        {"nace_codes.number": {"$regex": r"69101.*"}},
        {"establishments.nace_codes.number": {"$regex": r"69101.*"}},
        {"best_nace_codes.first_best_nace_code.number": {"$regex": r"69101.*"}},
        # Religious orgs
        {"nace_codes.number": {"$regex": r"9491.*"}},
        {"establishments.nace_codes.number": {"$regex": r"9491.*"}},
        {"best_nace_codes.first_best_nace_code.number": {"$regex": r"9491.*"}},
    ]
}
# Remove church factories (most of them doesn't have any declared NACE code)
NO_CHURCH_FACTORY = {
    "$expr": {
        "$not": {
            "$regexMatch": {
                "input": "$name",
                "regex": "communaute|gemeente|kerk|église|cathedrale|church factory|kirchenfabrik|eglise",
                "options": "i",
            }
        }
    }
}

# --- Geographic ----
COMPANIES_BE = {"country": Country.belgium.value}
COMPANIES_NL = {"country": Country.netherlands.value}
BE_FLANDER_REGION = {"address.region": RegionBe.flanders.value}
BE_WALLONIA_REGION = {"address.region": RegionBe.wallonia.value}

# --- Legal forms types ---
FOR_PROFIT_ONLY = {"legal_form_type": LegalFormCategory.for_profit}
NON_PROFIT_ONLY = {"legal_form_type": LegalFormCategory.non_profit}
PUBLIC_ONLY = {"legal_form_type": LegalFormCategory.public}
FOR_PROFIT_OR_PUBLIC = {"legal_form_type": {"$ne": LegalFormCategory.non_profit}}



import re
import idna
import asyncio
from typing import Literal
from tqdm.asyncio import tqdm
from aiodns import DNSResolver
from aiodns.error import DNSError
from pycares import ares_query_mx_result


DnsRecordType = Literal["A", "AAAA", "MX", "CNAME", "SRV", "TXT", "NS"]
# List of DNS servers to use for DNS resolution (Google & OpenDNS)
DNS_SERVERS = ["8.8.8.8", "208.67.222.222", "8.8.4.4", "208.67.220.220"]


def _validate_domain(domain: str) -> str | None:
    """
    Check the domain validity (fix it if possible).

    :return: the cleaned domain or None if it can't be fixed.
    """
    try:
        clean_domain = idna.encode(domain).decode("ascii").lower().strip()
    except (UnicodeError, idna.IDNAError):
        return None
    return clean_domain if re.match(r"^[\w.-]+\.\w{2,}$", clean_domain) else None


async def _check_domain_dns(domain: str, dns_types: list[DnsRecordType], resolver: DNSResolver) -> str | None:
    """Check if a domain has at least one of the dns_types DNS record."""
    validated_domain = _validate_domain(domain)
    if not validated_domain:
        return None
    for dns_type in dns_types:
        try:
            # Resolve the domain's DNS. Will raise an exception if the domain doesn't exist or timeout
            await resolver.query(validated_domain, dns_type)
            return domain
        # If one of these exceptions is raised, it means the domain doesn't exist or timed out
        except (DNSError, UnicodeError):
            continue
    return None


async def _get_domain_dns_mx_domain(domain: str, resolver: DNSResolver) -> tuple[str, str] | tuple[None, None]:
    """Return the domain's DNS record (based on the dns_type)."""
    validated_domain = _validate_domain(domain)
    if not validated_domain:
        return None, None
    try:
        records: list[ares_query_mx_result] = await resolver.query(domain, "MX")
        if not records:
            return None, None
        # get the first MX record host's domain
        mx_server_host = str(records[0].host)
        mx_server_domain = ".".join(mx_server_host.split(".")[-2:])
        return domain, mx_server_domain
    except (DNSError, UnicodeError):
        return None, None


async def check_if_domains_have_dns_async(
    domains: list[str],
    batch_size: int = 10_000,
    dns_timeout: float = 1,
    dns_types_to_check: list[DnsRecordType] = ["A", "AAAA"],
) -> set[str]:
    """Check all the domains in a wave. Return a set of valid domains."""
    valid_domains = []
    # Divide the domains in batches to avoid timeout issues
    domains_batches: list[list[str]] = [domains[i : i + batch_size] for i in range(0, len(domains), batch_size)]
    for domains_chunk in tqdm(domains_batches):
        resolver = DNSResolver(timeout=dns_timeout, nameservers=DNS_SERVERS, loop=asyncio.get_event_loop())
        tasks = [_check_domain_dns(domain, dns_types_to_check, resolver) for domain in domains_chunk]
        results = await asyncio.gather(*tasks)
        valid_domains.extend([r for r in results if r])
    return set(valid_domains)


def check_if_domains_have_dns(
    domains: list[str],
    batch_size: int = 10_000,
    dns_timeout: float = 1,
    dns_types_to_check: list[DnsRecordType] = ["A", "AAAA"],
) -> set[str]:
    """Check if a domain has at least one of the dns_types DNS record."""
    return asyncio.run(check_if_domains_have_dns_async(domains, batch_size, dns_timeout, dns_types_to_check))


async def get_domains_dns_mx_domain_async(
    domains: list[str], batch_size: int = 10_000, dns_timeout: float = 1
) -> dict[str, str]:
    """Check all the domains in a wave. Return a set of valid domains."""
    domain_to_mx_server_domain_mapper: dict[str, str] = {}
    # Divide the domains in batches to avoid timeout issues
    domains_batches: list[list[str]] = [domains[i : i + batch_size] for i in range(0, len(domains), batch_size)]
    for domains_chunk in tqdm(domains_batches):
        resolver = DNSResolver(timeout=dns_timeout, nameservers=DNS_SERVERS, loop=asyncio.get_event_loop())
        tasks = [_get_domain_dns_mx_domain(domain, resolver) for domain in domains_chunk]
        results = await asyncio.gather(*tasks)
        batch_mapper = {
            domain: mx_server_domain.strip().lower()  # MX server domain are registered in uppercase or with spaces
            for domain, mx_server_domain in results
            if mx_server_domain and domain
        }
        domain_to_mx_server_domain_mapper.update(batch_mapper)
    return domain_to_mx_server_domain_mapper


def get_domains_dns_mx_domain(domains: list[str], batch_size: int = 10_000, dns_timeout: float = 1) -> dict[str, str]:
    """Check all the domains in a wave. Return a set of valid domains."""
    return asyncio.run(get_domains_dns_mx_domain_async(domains, batch_size, dns_timeout))



"""
Allow import direcly from inoopa_utils without specifying the file (ie: from inoopa_utils import DbManagerMongo)
See: https://docs.astral.sh/ruff/rules/unused-import/
"""

# --- DB filter Mongo ---
from .common_mongo_filters import ACTIVE_COMPANIES as ACTIVE_COMPANIES
from .common_mongo_filters import BE_FLANDER_REGION as BE_FLANDER_REGION
from .common_mongo_filters import BE_WALLONIA_REGION as BE_WALLONIA_REGION
from .common_mongo_filters import COMPANIES_BE as COMPANIES_BE
from .common_mongo_filters import COMPANIES_NL as COMPANIES_NL
from .common_mongo_filters import FOR_PROFIT_ONLY as FOR_PROFIT_ONLY
from .common_mongo_filters import FOR_PROFIT_OR_PUBLIC as FOR_PROFIT_OR_PUBLIC
from .common_mongo_filters import NO_CHURCH_FACTORY as NO_CHURCH_FACTORY
from .common_mongo_filters import NO_LAWYER_NO_RELGIOUS_ORGS as NO_LAWYER_NO_RELGIOUS_ORGS
from .common_mongo_filters import NON_PROFIT_ONLY as NON_PROFIT_ONLY
from .common_mongo_filters import PUBLIC_ONLY as PUBLIC_ONLY
from .common_mongo_filters import WITH_PHONE as WITH_PHONE
from .common_mongo_filters import WITH_PHONE_NO_DNCM as WITH_PHONE_NO_DNCM
from .common_mongo_filters import WITH_WEBSITE as WITH_WEBSITE

# --- Helpers ---
from .helpers import extract_domain as extract_domain
from .helpers import get_all_latest_user_agents as get_all_latest_user_agents
from .helpers import get_latest_user_agent as get_latest_user_agent
from .helpers import get_random_user_agent as get_random_user_agent

# --- Logging ---
from .inoopa_logging import create_logger as create_logger

# --- DB ---
from .mongodb_helpers import DbManagerMongo as DbManagerMongo
from .typesense_helpers import TypesenseManager as TypesenseManager



from dataclasses import dataclass
from datetime import datetime

from inoopa_utils.custom_types.addresses import Country

# Map employee_categories codes with their labels
EMPLOYEE_CATEGORIES_CODES_MAP: dict[int | None, str] = {
    0: "emp_0",
    1: "emp_1_to_4",
    2: "emp_5_to_9",
    3: "emp_10_to_19",
    4: "emp_20_to_49",
    5: "emp_50_to_99",
    6: "emp_100_to_199",
    7: "emp_200_to_499",
    8: "emp_500_to_999",
    9: "emp_1000_plus",
    None: "unknown",
}


@dataclass(slots=True)
class PerEmployeeCategory:
    """Hold the counting of a metric."""

    total: int | None = None
    distinct: int | None = None
    emp_0: int = 0
    emp_1_to_4: int = 0
    emp_5_to_9: int = 0
    emp_10_to_19: int = 0
    emp_20_to_49: int = 0
    emp_50_to_99: int = 0
    emp_100_to_199: int = 0
    emp_200_to_499: int = 0
    emp_500_to_999: int = 0
    emp_1000_plus: int = 0
    unknown: int = 0


@dataclass(slots=True)
class CompanyStats:
    """
    Hold the metrics of our company collection at a given point in time.
    Each metric is divided by employee category.
    """

    date: datetime
    country: Country
    active_companies: PerEmployeeCategory
    websites: PerEmployeeCategory
    emails: PerEmployeeCategory
    phones: PerEmployeeCategory
    attributed_phones: PerEmployeeCategory


@dataclass(slots=True)
class DecisionMakersStats:
    """
    Hold the stats of the decision_makers collection at a given point in time.
    Each metric is divided by employee category.
    Each metric has a _dm version that counts unique decision makers.
    Each metric has a _companies version that counts unique companies with at least 1 matching DM.
    """

    date: datetime
    country: Country
    with_name_dms: PerEmployeeCategory
    with_job_title_dms: PerEmployeeCategory
    with_department_dms: PerEmployeeCategory
    with_responsibility_level_dms: PerEmployeeCategory
    with_linkedin_url_dms: PerEmployeeCategory
    with_email_dms: PerEmployeeCategory
    with_name_companies: PerEmployeeCategory
    with_job_title_companies: PerEmployeeCategory
    with_department_companies: PerEmployeeCategory
    with_responsibility_level_companies: PerEmployeeCategory
    with_linkedin_url_companies: PerEmployeeCategory
    with_email_companies: PerEmployeeCategory
    board_members: PerEmployeeCategory
    board_members_companies: PerEmployeeCategory


@dataclass(slots=True)
class DecisionMakersDepartmentStats:
    """
    Hold the stats of the decision_makers collection per department at a given point in time.
    Each metric is divided by employee category.
    Each metric has a _dm version that counts unique decision makers.
    """

    date: datetime
    country: Country
    from_it_dms: PerEmployeeCategory
    from_it_dms_with_emails: PerEmployeeCategory
    from_sales_dms: PerEmployeeCategory
    from_sales_dms_with_emails: PerEmployeeCategory
    from_legal_dms: PerEmployeeCategory
    from_legal_dms_with_emails: PerEmployeeCategory
    from_finance_dms: PerEmployeeCategory
    from_finance_dms_with_emails: PerEmployeeCategory
    from_board_dms: PerEmployeeCategory
    from_board_dms_with_emails: PerEmployeeCategory
    from_logistic_dms: PerEmployeeCategory
    from_logistic_dms_with_emails: PerEmployeeCategory
    from_hr_dms: PerEmployeeCategory
    from_hr_dms_with_emails: PerEmployeeCategory
    from_support_dms: PerEmployeeCategory
    from_support_dms_with_emails: PerEmployeeCategory
    from_esg_dms: PerEmployeeCategory
    from_esg_dms_with_emails: PerEmployeeCategory
    from_administration_dms: PerEmployeeCategory
    from_administration_dms_with_emails: PerEmployeeCategory
    from_project_management_dms: PerEmployeeCategory
    from_project_management_dms_with_emails: PerEmployeeCategory
    from_notworking_dms: PerEmployeeCategory
    from_notworking_dms_with_emails: PerEmployeeCategory
    from_other_dms: PerEmployeeCategory
    from_other_dms_with_emails: PerEmployeeCategory


@dataclass(slots=True)
class NaceMetricForEmployeeCategory:
    """
    Hold the distribution of companies per sector (lvl 1 NACE) or per code (any lvl)

    Both will get a dict like: {nace_level: number_of_companies}
    ex: {"12": 333} -> "12" is the nace code, "333" is the number of companies having this code as first best NACE
    """

    per_sector: dict[str, int]  # Level 1
    per_code: dict[str, int]  # Any level


@dataclass(slots=True)
class NaceStats:
    """
    Hold the metrics of country's NACE.
    Each metric is divided by employee category.
    """

    date: datetime
    country: Country
    total: NaceMetricForEmployeeCategory
    emp_0: NaceMetricForEmployeeCategory
    emp_1_to_4: NaceMetricForEmployeeCategory
    emp_5_to_9: NaceMetricForEmployeeCategory
    emp_10_to_19: NaceMetricForEmployeeCategory
    emp_20_to_49: NaceMetricForEmployeeCategory
    emp_50_to_99: NaceMetricForEmployeeCategory
    emp_100_to_199: NaceMetricForEmployeeCategory
    emp_200_to_499: NaceMetricForEmployeeCategory
    emp_500_to_999: NaceMetricForEmployeeCategory
    emp_1000_plus: NaceMetricForEmployeeCategory


def dict_to_nace_stats(nace_stats: dict) -> NaceStats:
    """Convert a dict from the DB to a NaceStats dataclass."""
    return NaceStats(
        date=nace_stats["date"],
        country=Country(nace_stats["country"]),
        total=NaceMetricForEmployeeCategory(**nace_stats["total"]),
        emp_0=NaceMetricForEmployeeCategory(**nace_stats["emp_0"]),
        emp_1_to_4=NaceMetricForEmployeeCategory(**nace_stats["emp_1_to_4"]),
        emp_5_to_9=NaceMetricForEmployeeCategory(**nace_stats["emp_5_to_9"]),
        emp_10_to_19=NaceMetricForEmployeeCategory(**nace_stats["emp_10_to_19"]),
        emp_20_to_49=NaceMetricForEmployeeCategory(**nace_stats["emp_20_to_49"]),
        emp_50_to_99=NaceMetricForEmployeeCategory(**nace_stats["emp_50_to_99"]),
        emp_100_to_199=NaceMetricForEmployeeCategory(**nace_stats["emp_100_to_199"]),
        emp_200_to_499=NaceMetricForEmployeeCategory(**nace_stats["emp_200_to_499"]),
        emp_500_to_999=NaceMetricForEmployeeCategory(**nace_stats["emp_500_to_999"]),
        emp_1000_plus=NaceMetricForEmployeeCategory(**nace_stats["emp_1000_plus"]),
    )


def dict_to_company_stats(company_stats: dict) -> CompanyStats:
    """Convert a dict from the DB to a CompanyStats dataclass."""
    company_stats_fmt = CompanyStats(
        date=company_stats["date"],
        country=Country(company_stats["country"]),
        active_companies=PerEmployeeCategory(**company_stats["active_companies"]),
        websites=PerEmployeeCategory(**company_stats["websites"]),
        phones=PerEmployeeCategory(**company_stats["phones"]),
        emails=PerEmployeeCategory(**company_stats["emails"]),
        attributed_phones=PerEmployeeCategory(**company_stats["attributed_phones"]),
    )
    return company_stats_fmt


def dict_to_decision_makers_stats(decision_makers_stats: dict) -> DecisionMakersStats:
    """Convert a dict from the DB to a DecisionMakersStats dataclass."""
    decision_makers_stats_fmt = DecisionMakersStats(
        date=decision_makers_stats["date"],
        country=Country(decision_makers_stats["country"]),
        with_name_dms=PerEmployeeCategory(**decision_makers_stats["with_name_dms"]),
        with_job_title_dms=PerEmployeeCategory(**decision_makers_stats["with_job_title_dms"]),
        with_department_dms=PerEmployeeCategory(**decision_makers_stats["with_department_dms"]),
        with_responsibility_level_dms=PerEmployeeCategory(**decision_makers_stats["with_responsibility_level_dms"]),
        with_linkedin_url_dms=PerEmployeeCategory(**decision_makers_stats["with_linkedin_url_dms"]),
        with_email_dms=PerEmployeeCategory(**decision_makers_stats["with_email_dms"]),
        with_name_companies=PerEmployeeCategory(**decision_makers_stats["with_name_companies"]),
        with_job_title_companies=PerEmployeeCategory(**decision_makers_stats["with_job_title_companies"]),
        with_department_companies=PerEmployeeCategory(**decision_makers_stats["with_department_companies"]),
        with_responsibility_level_companies=PerEmployeeCategory(
            **decision_makers_stats["with_responsibility_level_companies"]
        ),
        with_linkedin_url_companies=PerEmployeeCategory(**decision_makers_stats["with_linkedin_url_companies"]),
        with_email_companies=PerEmployeeCategory(**decision_makers_stats["with_email_companies"]),
        board_members=PerEmployeeCategory(**decision_makers_stats["board_members"]),
        board_members_companies=PerEmployeeCategory(**decision_makers_stats["board_members_companies"]),
    )
    return decision_makers_stats_fmt


def dict_to_decision_makers_departments_stats(decision_makers_stats: dict) -> DecisionMakersDepartmentStats:
    """Convert a dict from the DB to a DecisionMakersDepartmentStats dataclass."""
    decision_makers_stats_fmt = DecisionMakersDepartmentStats(
        date=decision_makers_stats["date"],
        country=Country(decision_makers_stats["country"]),
        from_it_dms=PerEmployeeCategory(**decision_makers_stats["from_it_dms"]),
        from_it_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_it_dms_with_emails"]),
        from_sales_dms=PerEmployeeCategory(**decision_makers_stats["from_sales_dms"]),
        from_sales_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_sales_dms_with_emails"]),
        from_legal_dms=PerEmployeeCategory(**decision_makers_stats["from_legal_dms"]),
        from_legal_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_legal_dms_with_emails"]),
        from_finance_dms=PerEmployeeCategory(**decision_makers_stats["from_finance_dms"]),
        from_finance_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_finance_dms_with_emails"]),
        from_board_dms=PerEmployeeCategory(**decision_makers_stats["from_board_dms"]),
        from_board_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_board_dms_with_emails"]),
        from_logistic_dms=PerEmployeeCategory(**decision_makers_stats["from_logistic_dms"]),
        from_logistic_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_logistic_dms_with_emails"]),
        from_hr_dms=PerEmployeeCategory(**decision_makers_stats["from_hr_dms"]),
        from_hr_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_hr_dms_with_emails"]),
        from_support_dms=PerEmployeeCategory(**decision_makers_stats["from_support_dms"]),
        from_support_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_support_dms_with_emails"]),
        from_esg_dms=PerEmployeeCategory(**decision_makers_stats["from_esg_dms"]),
        from_esg_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_esg_dms_with_emails"]),
        from_notworking_dms=PerEmployeeCategory(**decision_makers_stats["from_notworking_dms"]),
        from_notworking_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_notworking_dms_with_emails"]),
        from_other_dms=PerEmployeeCategory(**decision_makers_stats["from_other_dms"]),
        from_other_dms_with_emails=PerEmployeeCategory(**decision_makers_stats["from_other_dms_with_emails"]),
        from_administration_dms=PerEmployeeCategory(**decision_makers_stats["from_administration_dms"]),
        from_administration_dms_with_emails=PerEmployeeCategory(
            **decision_makers_stats["from_administration_dms_with_emails"]
        ),
        from_project_management_dms=PerEmployeeCategory(**decision_makers_stats["from_project_management_dms"]),
        from_project_management_dms_with_emails=PerEmployeeCategory(
            **decision_makers_stats["from_project_management_dms_with_emails"]
        ),
    )
    return decision_makers_stats_fmt



"""Base Inoopa configuration for python Logging using Logfire."""

import os
from typing import Literal

import logfire
from dotenv import load_dotenv

from inoopa_utils.utils.env_variables_helper import get_env_name

load_dotenv()


LoggingLevel = Literal["CRITICAL", "FATAL", "ERROR", "WARNING", "INFO", "DEBUG"]


def create_logger(
    logger_name: str, logging_level: LoggingLevel | None = None, logs_dir_path: str = "./logs", pretty: bool = False
) -> logfire.Logfire:
    """
    Configure how logging should be done using Logfire.

    :param logger_name: The logger name to return.
    :param logging_level: The level of logging to filter. If none, will deduce from "ENV" env variable:
        'dev' will set logging_level to "DEBUG"
        'staging' will set logging_level to "INFO"
        'prod' will set logging_level to "INFO"
    :param logs_dir_path: The path to the logs directory.
    :param pretty: If True, will use rich to pretty print the logs. Only for development & CLI purpose.
    """
    env = os.getenv("ENV", "LOCAL")
    if logging_level is None:
        logging_level = "DEBUG" if get_env_name() == "dev" else "INFO"

    # Convert logging level to logfire level
    logfire_level = logging_level.lower()

    # Configure logfire with basic settings
    console_options = logfire.ConsoleOptions(
        min_log_level=logfire_level,  # type: ignore
        colors="auto",
        include_timestamps=True,
        verbose=False,
        show_project_link=False,
        include_tags=True,
        span_style="indented",
    )

    logfire.configure(
        service_name=logger_name,
        console=console_options,
        send_to_logfire="if-token-present",  # Only send to Logfire platform if token is available
        min_level=logfire_level,  # type: ignore
        environment=env,
        scrubbing=logfire.ScrubbingOptions(callback=scrubbing_callback),
    )

    # Create a logger instance with the service name as a tag
    logger = logfire.with_settings(tags=[logger_name, env])

    return logger


def scrubbing_callback(m: logfire.ScrubMatch):
    """Prevent Logfire to redact Strip error messages which are needed to understand what went wrong."""
    if m.path == ("message", "e") and m.pattern_match.group(0) == "Session":
        return m.value
    if m.path == ("attributes", "e") and m.pattern_match.group(0) == "Session":
        return m.value


if __name__ == "__main__":
    logger = create_logger("test_logger", logging_level="DEBUG", pretty=True)
    logger.debug("Debug message")
    logger.info("Info message with {data}", data={"hello": "world"})
    logger.warning("Warning message")
    logger.error("Error message")



"""
Allow import direcly from inoopa_utils.custom_types without specifying file (ie: from inoopa_utils.custom_types import Company)
See: https://docs.astral.sh/ruff/rules/unused-import/
"""

# --- Address ---
from .addresses import Country as Country
from .addresses import ProvinceBe as ProvinceBe
from .addresses import ProvinceFr as ProvinceFr
from .addresses import ProvinceNl as ProvinceNl
from .addresses import RegionBe as RegionBe
from .addresses import RegionFr as RegionFr
from .addresses import RegionNl as RegionNl

# --- Company ---
from .companies import Address as Address
from .companies import BestNaceCodes as BestNaceCodes
from .companies import BoardMember as BoardMember
from .companies import Company as Company
from .companies import Email as Email
from .companies import Establishment as Establishment
from .companies import Finance as Finance
from .companies import NaceCode as NaceCode
from .companies import NaceCodeExtended as NaceCodeExtended
from .companies import Phone as Phone
from .companies import Website as Website
from .companies import convert_dict_to_company as convert_dict_to_company

# --- DB stats ---
from .db_stats import EMPLOYEE_CATEGORIES_CODES_MAP as EMPLOYEE_CATEGORIES_CODES_MAP
from .db_stats import CompanyStats as CompanyStats
from .db_stats import DecisionMakersStats as DecisionMakersStats
from .db_stats import NaceMetricForEmployeeCategory as NaceMetricForEmployeeCategory
from .db_stats import NaceStats as NaceStats
from .db_stats import PerEmployeeCategory as PerEmployeeCategory
from .db_stats import dict_to_company_stats as dict_to_company_stats
from .db_stats import dict_to_decision_makers_stats as dict_to_decision_makers_stats
from .db_stats import dict_to_nace_stats as dict_to_nace_stats

# --- Decision Makers ---
from .decision_makers import DecisionMaker as DecisionMaker
from .decision_makers import DecisionMakerDepartment as DecisionMakerDepartment
from .decision_makers import responsibilities_level_mapper as responsibilities_level_mapper
from .delivery_filters import AdditionalFields as AdditionalFields
from .delivery_filters import EnrichmentCompanyFilters as EnrichmentCompanyFilters
from .delivery_filters import LeadGenerationCompanyFilters as LeadGenerationCompanyFilters

# --- Delivery filters ---
from .delivery_filters import MongoFilters as MongoFilters

# --- Legal filters ---
from .legal_filters import EmployeeCategoryClass as EmployeeCategoryClass
from .legal_filters import EntityType as EntityType
from .legal_filters import LegalFormBe as LegalFormBe
from .legal_filters import LegalFormCategory as LegalFormCategory
from .legal_filters import LegalFormNl as LegalFormNl
from .legal_filters import get_all_legal_forms as get_all_legal_forms
from .legal_filters import get_all_legal_forms_str as get_all_legal_forms_str

# --- Website structure ---
from .website_structure import WebsiteStructure as WebsiteStructure



import json
from datetime import datetime
from typing import Literal

from pydantic import BaseModel

from inoopa_utils.custom_types import DecisionMakerDepartment
from inoopa_utils.custom_types.addresses import (
    Country,
    ProvinceBe,
    ProvinceFr,
    ProvinceNl,
    RegionBe,
    RegionFr,
    RegionNl,
)
from inoopa_utils.custom_types.legal_filters import (
    EmployeeCategoryClass,
    EntityType,
    LegalFormBe,
    LegalFormCategory,
    LegalFormNl,
)

MongoFilters = dict[str, dict[str, dict[str, int | str | list[str]]]]

AdditionalFields = Literal[
    "email",
    "phone",
    "website",
    "nace_codes",
    "social_medias",
    "decision_makers_name",
    "decision_makers_email",
    "board_members",
]


class LeadGenerationCompanyFilters(BaseModel):
    """
    Represents the filters used to generate the leads.

    This is a pydantic model, so it can be used in the apis as a parameter.
    """

    additional_fields: list[AdditionalFields] | None = [
        "email",
        "phone",
        "website",
        "nace_codes",
        "social_medias",
        "decision_makers_name",
        "decision_makers_email",
        "board_members",
    ]

    countries: list[Country] | None = None
    regions: list[RegionBe | RegionFr | RegionNl] | None = None
    provinces: list[ProvinceBe | ProvinceFr | ProvinceNl] | None = None
    zipcodes: list[str] | None = None
    declared_nace_codes: list[str] | None = None
    declared_nace_codes_inclusive: list[str] | None = None
    best_nace_codes: list[str] | None = None
    best_nace_codes_inclusive: list[str] | None = None

    minimum_number_of_establishments: int | None = None
    maximum_number_of_establishments: int | None = None
    employee_categories: list[EmployeeCategoryClass] | None = None
    created_before: datetime | None = None
    created_after: datetime | None = None

    max_results: int | None = None
    max_decision_makers_per_company: int | None = None
    decision_makers_department_allowed: list[DecisionMakerDepartment] | None = None
    decision_makers_department_priority: list[DecisionMakerDepartment] | None = None
    excluded_companies: list[str] | None = None

    legal_form_Categories: list[LegalFormCategory] | None = None
    legal_forms: list[LegalFormBe | LegalFormNl] | None = None
    entity_types: list[EntityType] | None = None

    only_include_active_companies: bool = True
    only_include_companies_with_phone_no_do_not_call_me: bool = False
    only_include_companies_with_email: bool = False
    only_include_companies_with_website: bool = False
    only_include_companies_with_nace_codes: bool = False

    def to_mongo_filters(self) -> MongoFilters:
        """
        Generate a MongoDB filter to search for companies based on the LeadGenerationCompanyFilters.

        can be used as a filter parameter in the MongoDB Collection.find() method.
        """
        filters = {}
        if self.countries:
            filters["country"] = {"$in": [c.value for c in self.countries]}
        if self.legal_forms:
            filters["legal_form"] = {"$in": [form.value for form in self.legal_forms]}
        if self.employee_categories:
            filters["employee_category_formatted"] = {"$in": [c.value for c in self.employee_categories]}
        if self.regions:
            regions = [r.value if r.value != RegionBe.not_found.value else None for r in self.regions]
            filters["address.region"] = {"$in": regions}
        if self.provinces:
            provinces = [p.value if p.value != ProvinceBe.not_found.value else None for p in self.provinces]
            filters["address.province"] = {"$in": provinces}
        if self.zipcodes:
            zipcodes = [z if z != "NOT FOUND" else None for z in self.zipcodes]
            filters["address.postal_code"] = {"$in": zipcodes}

        if self.minimum_number_of_establishments and self.maximum_number_of_establishments:
            filters["number_of_establishments"] = {
                "$gte": self.minimum_number_of_establishments,
                "$lte": self.maximum_number_of_establishments,
            }
        elif self.minimum_number_of_establishments:
            filters["number_of_establishments"] = {"$gte": self.minimum_number_of_establishments}
        elif self.maximum_number_of_establishments:
            filters["number_of_establishments"] = {"$lte": self.maximum_number_of_establishments}

        if self.declared_nace_codes:
            filters["$or"] = [
                {"nace_codes": {"$elemMatch": {"number": {"$in": self.declared_nace_codes}}}},
                {
                    "establishments": {
                        "$elemMatch": {"nace_codes": {"$elemMatch": {"number": {"$in": self.declared_nace_codes}}}}
                    }
                },
            ]
        if self.declared_nace_codes_inclusive:
            nace_section_codes = []
            nace_codes_regex = []
            for code in self.declared_nace_codes_inclusive:
                # if the regex is only one letter, it's a nace section code
                if len(code) == 1:
                    # If the section code is not a letter, something went wrong, ignore it
                    if not code.isalpha():
                        continue
                    nace_section_codes.append(code.upper())
                else:
                    # if the regex is more than one letter, it's a nace code regex
                    nace_codes_regex.append(f"^{code}.*")

            filters["$or"] = []
            if nace_codes_regex:
                filters["$or"].append(
                    {"nace_codes": {"$elemMatch": {"number": {"$regex": "|".join(nace_codes_regex)}}}}
                )
                filters["$or"].append(
                    {
                        "establishments": {
                            "$elemMatch": {
                                "nace_codes": {"$elemMatch": {"number": {"$regex": "|".join(nace_codes_regex)}}}
                            }
                        }
                    },
                )
            if nace_section_codes:
                filters["$or"].append({"nace_codes": {"$elemMatch": {"section_code": {"$in": nace_section_codes}}}})
                filters["$or"].append(
                    {
                        "establishments": {
                            "$elemMatch": {"nace_codes": {"$elemMatch": {"section_code": {"$in": nace_section_codes}}}}
                        }
                    },
                )

        if self.best_nace_codes:
            filters["best_nace_codes.first_best_nace_code.number"] = {"$in": self.best_nace_codes}
        if self.best_nace_codes_inclusive:
            nace_codes = []
            nace_section_codes = []
            for code in self.best_nace_codes_inclusive:
                if len(code) == 1:
                    if not code.isalpha():
                        continue
                    nace_section_codes.append(code.upper())
                else:
                    # if the code is more than one letter, it's a nace code not a section
                    nace_codes.append(f"^{code}.*")
            filters["$or"] = []
            if nace_codes:
                filters["$or"].append({"best_nace_codes.first_best_nace_code.number": {"$regex": "|".join(nace_codes)}})
            if nace_section_codes:
                filters["$or"].append(
                    {"best_nace_codes.first_best_nace_code.section_code": {"$in": nace_section_codes}}
                )

        if self.legal_form_Categories:
            filters["legal_form_type"] = {"$in": [c.value for c in self.legal_form_Categories]}
        if self.entity_types:
            filters["entity_type"] = {"$in": [e.value for e in self.entity_types]}
        if self.created_before and not self.created_after:
            filters["start_date"] = {"$lte": self.created_before}
        if self.created_after and not self.created_before:
            filters["start_date"] = {"$gte": self.created_after}
        if self.created_before and self.created_after:
            filters["start_date"] = {"$lte": self.created_before, "$gte": self.created_after}
        if self.excluded_companies:
            filters["_id"] = {"$nin": self.excluded_companies}
        if self.only_include_active_companies:
            filters["status"] = "Active"
            filters["legal_situation"] = "Normal situation"

        if self.only_include_companies_with_phone_no_do_not_call_me:
            filters["best_phone"] = {"$ne": None}
            filters["best_phone.phone"] = {"$nin": ["DO_NOT_CALL_ME", None]}

        if self.only_include_companies_with_email:
            filters["best_email"] = {"$ne": None}
        if self.only_include_companies_with_website:
            filters["best_website"] = {"$ne": None}
        if self.only_include_companies_with_nace_codes:
            filters["best_nace_codes"] = {"$ne": None}
        return _optimize_query(filters)

    def to_dict(self) -> dict:
        data_dict = self.model_dump()
        # Convert datetime objects to isoformat for json serialization
        if self.created_before:
            data_dict["created_before"] = _datetime_serializer(self.created_before)
        if self.created_after:
            data_dict["created_after"] = _datetime_serializer(self.created_after)
        return data_dict

    def to_hash(self, semantic_search_query: str | None = None) -> int:
        """Hash the filters to allow comparing."""
        hashable_dict = {
            "minimum_number_of_establishments": self.minimum_number_of_establishments,
            "maximum_number_of_establishments": self.maximum_number_of_establishments,
            "max_results": self.max_results,
            "created_before": self.created_before.__str__() if self.created_before else None,
            "created_after": self.created_after.__str__() if self.created_after else None,
            "max_decision_makers_per_company": self.max_decision_makers_per_company,
            # Sort all lists to get a similar hash for similar requests
            "zipcodes": sorted(self.zipcodes) if self.zipcodes else None,
            "declared_nace_codes": sorted(self.declared_nace_codes) if self.declared_nace_codes else None,
            "declared_nace_codes_inclusive": sorted(self.declared_nace_codes_inclusive)
            if self.declared_nace_codes_inclusive
            else None,
            "best_nace_codes": sorted(self.best_nace_codes) if self.best_nace_codes else None,
            "best_nace_codes_inclusive": sorted(self.best_nace_codes_inclusive)
            if self.best_nace_codes_inclusive
            else None,
            "excluded_companies": sorted(self.excluded_companies) if self.excluded_companies else None,
            "countries": sorted([v.value for v in self.countries]) if self.countries else None,
            "employee_categories": sorted([v.value for v in self.employee_categories])
            if self.employee_categories
            else None,
            "legal_form_Categories": sorted([v.value for v in self.legal_form_Categories])
            if self.legal_form_Categories
            else None,
            "entity_types": sorted([v.value for v in self.entity_types]) if self.entity_types else None,
            "legal_forms": sorted([v.value for v in self.legal_forms]) if self.legal_forms else None,
            "regions": sorted([v.value for v in self.regions]) if self.regions else None,
            "provinces": sorted([v.value for v in self.provinces]) if self.provinces else None,
            "decision_makers_department_allowed": sorted([v.value for v in self.decision_makers_department_allowed])
            if self.decision_makers_department_allowed
            else None,
            # ! This one requires following the user's order
            "decision_makers_department_priority": [v.value for v in self.decision_makers_department_priority]
            if self.decision_makers_department_priority
            else None,
            "only_include_active_companies": self.only_include_active_companies,
            "only_include_companies_with_phone_no_do_not_call_me": self.only_include_companies_with_phone_no_do_not_call_me,
            "only_include_companies_with_email": self.only_include_companies_with_email,
            "only_include_companies_with_website": self.only_include_companies_with_website,
            "only_include_companies_with_nace_codes": self.only_include_companies_with_nace_codes,
        }
        dict_str_dump = json.dumps(hashable_dict, sort_keys=True, default=str, ensure_ascii=False)
        if semantic_search_query:
            dict_str_dump += semantic_search_query.strip().lower()
        return hash(dict_str_dump)


class EnrichmentCompanyFilters(BaseModel):
    vats_to_enrich: list[str]

    only_include_active_companies: bool = True
    only_include_companies_with_phone_no_do_not_call_me: bool = False
    only_include_companies_with_email: bool = False
    only_include_companies_with_website: bool = False
    only_include_companies_with_nace_codes: bool = False

    additional_fields: list[AdditionalFields] = [
        "email",
        "phone",
        "website",
        "nace_codes",
        "social_medias",
        "decision_makers_name",
        "decision_makers_email",
        "board_members",
    ]

    def to_mongo_filters(self) -> MongoFilters:
        filters: dict = {"_id": {"$in": self.vats_to_enrich}}
        if self.only_include_active_companies:
            filters["status"] = "Active"
            filters["legal_situation"] = "Normal situation"
        if self.only_include_companies_with_phone_no_do_not_call_me:
            filters["best_phone"] = {"$ne": None}
        if self.only_include_companies_with_email:
            filters["best_email"] = {"$ne": None}
        if self.only_include_companies_with_website:
            filters["best_website"] = {"$ne": None}
        if self.only_include_companies_with_nace_codes:
            filters["best_nace_codes"] = {"$ne": None}
        return _optimize_query(filters)

    def to_dict(self) -> dict:
        return self.model_dump()

    def to_hash(self) -> int:
        """Hash the filters to allow comparing."""
        dict_str_dump = json.dumps(self.model_dump(), sort_keys=True, default=str, ensure_ascii=False)
        return hash(dict_str_dump)


def _datetime_serializer(obj):
    if isinstance(obj, datetime):
        return obj.isoformat()
    raise TypeError(f"Type {type(obj)} not serializable")


def _optimize_query(filters: dict) -> dict:
    """
    Recursively optimize MongoDB filters by simplifying single-element $in/$nin/$or operators.

    Transformations:
    - {"$in": [single_value]} → single_value (direct equality)
    - {"$nin": [single_value]} → {"$ne": single_value}
    - {"$or": [single_condition]} → single_condition (inlined into parent)

    This preserves the structure of complex queries with $or, $elemMatch, etc.
    """
    if not filters:
        return filters

    result = {}
    for key, value in filters.items():
        if isinstance(value, dict):
            # Check for single-element $in with only $in key
            if "$in" in value and len(value) == 1 and isinstance(value["$in"], list) and len(value["$in"]) == 1:
                result[key] = value["$in"][0]
            # Check for single-element $nin with only $nin key
            elif "$nin" in value and len(value) == 1 and isinstance(value["$nin"], list) and len(value["$nin"]) == 1:
                result[key] = {"$ne": value["$nin"][0]}
            else:
                # Recursively process nested dicts
                result[key] = _optimize_query(value)
        elif isinstance(value, list):
            # Single-element $or: inline the condition into the parent dict
            if key == "$or" and len(value) == 1 and isinstance(value[0], dict):
                optimized = _optimize_query(value[0])
                result.update(optimized)
            else:
                # Recursively process lists (for $or arrays, etc.)
                result[key] = [_optimize_query(item) if isinstance(item, dict) else item for item in value]
        else:
            result[key] = value

    return result



from enum import Enum


class ProvinceBe(Enum):
    antwerpen = "Antwerpen (BE)"
    vlaams_brabant = "Vlaams-Brabant (BE)"
    brabant_wallon = "Brabant-Wallon (BE)"
    bruxelles = "Bruxelles (BE)"
    west_vlaanderen = "West-Vlaanderen (BE)"
    oost_vlaanderen = "Oost-Vlaanderen (BE)"
    hainaut = "Hainaut (BE)"
    limbourg = "Limbourg (BE)"
    liege = "Liege (BE)"
    luxembourg = "Luxembourg (BE)"
    namur = "Namur (BE)"
    not_found = "NOT DECLARED"

    @classmethod
    def get_all_values(cls):
        """Return all possible values for this enum as a list of enum."""
        return [v for v in cls.__dict__.values() if isinstance(v, cls)]

    @classmethod
    def get_all_values_str(cls) -> list[str]:
        """Return all possible values for this enum as a list of str."""
        return [v.value for v in cls.__dict__.values() if isinstance(v, cls)]


class ProvinceFr(Enum):
    @classmethod
    def __new__(cls, value):
        raise NotImplementedError("FR provinces not implemented yet")

    @classmethod
    def get_all_values(cls):
        """Return all possible values for this enum as a list of enum."""
        return [v for v in cls.__dict__.values() if isinstance(v, cls)]

    @classmethod
    def get_all_values_str(cls) -> list[str]:
        """Return all possible values for this enum as a list of str."""
        return [v.value for v in cls.__dict__.values() if isinstance(v, cls)]


class ProvinceNl(Enum):
    noord_holland = "Noord-Holland (NL)"
    zuid_holland = "Zuid-Holland (NL)"
    utrecht = "Utrecht (NL)"
    zeeland = "Zeeland (NL)"
    noord_brabant = "Noord-Brabant (NL)"
    limburg = "Limburg (NL)"
    gelderland = "Gelderland (NL)"
    overijssel = "Overijssel (NL)"
    drenthe = "Drenthe (NL)"
    flevoland = "Flevoland (NL)"
    friesland = "Friesland (NL)"
    groningen = "Groningen (NL)"
    not_found = "NOT DECLARED"

    @classmethod
    def get_all_values(cls):
        """Return all possible values for this enum as a list of enum."""
        return [v for v in cls.__dict__.values() if isinstance(v, cls)]

    @classmethod
    def get_all_values_str(cls) -> list[str]:
        """Return all possible values for this enum as a list of str."""
        return [v.value for v in cls.__dict__.values() if isinstance(v, cls)]


class RegionBe(Enum):
    wallonia = "Wallonie (BE)"
    flanders = "Vlaanderen (BE)"
    bruxelles = "Bruxelles (BE)"
    not_found = "NOT DECLARED"

    @classmethod
    def get_all_values(cls):
        """Return all possible values for this enum as a list of enum."""
        return [v for v in cls.__dict__.values() if isinstance(v, cls)]

    @classmethod
    def get_all_values_str(cls) -> list[str]:
        """Return all possible values for this enum as a list of str."""
        return [v.value for v in cls.__dict__.values() if isinstance(v, cls)]


class RegionFr(Enum):
    @classmethod
    def __new__(cls, value):
        raise NotImplementedError("FR regions not implemented yet")

    @classmethod
    def get_all_values(cls):
        """Return all possible values for this enum as a list of enum."""
        return [v for v in cls.__dict__.values() if isinstance(v, cls)]

    @classmethod
    def get_all_values_str(cls) -> list[str]:
        """Return all possible values for this enum as a list of str."""
        return [v.value for v in cls.__dict__.values() if isinstance(v, cls)]


class RegionNl(Enum):
    western_netherlands = "Western Netherlands (NL)"
    northern_netherlands = "Northern Netherlands (NL)"
    eastern_netherlands = "Eastern Netherlands (NL)"
    southern_netherlands = "Southern Netherlands (NL)"
    not_found = "NOT DECLARED"

    @classmethod
    def get_all_values(cls):
        """Return all possible values for this enum as a list of enum."""
        return [v for v in cls.__dict__.values() if isinstance(v, cls)]

    @classmethod
    def get_all_values_str(cls) -> list[str]:
        """Return all possible values for this enum as a list of str."""
        return [v.value for v in cls.__dict__.values() if isinstance(v, cls)]


class Country(Enum):
    belgium = "BE"
    france = "FR"
    netherlands = "NL"

    @classmethod
    def get_all_values(cls):
        """Return all possible values for this enum as a list of enum."""
        return [v for v in cls.__dict__.values() if isinstance(v, cls)]

    @classmethod
    def get_all_values_str(cls) -> list[str]:
        """Return all possible values for this enum as a list of str."""
        return [v.value for v in cls.__dict__.values() if isinstance(v, cls)]


Region = RegionBe | RegionFr | RegionNl
Province = ProvinceBe | ProvinceFr | ProvinceNl


def get_all_regions(countries: list[Country] | list[str]) -> list[Region]:
    """Get all regions for a list of countries. Countires can be a list of Country enums or a list of str."""
    all_regions = []
    if not countries:
        return all_regions

    if isinstance(countries[0], str):
        countries = [Country(country) for country in countries]

    if Country.belgium in countries:
        all_regions.extend(RegionBe.get_all_values())
    if Country.france in countries:
        all_regions.extend(RegionFr.get_all_values())
    if Country.netherlands in countries:
        all_regions.extend(RegionNl.get_all_values())
    return all_regions


def get_all_regions_str(countries: list[Country] | list[str]) -> list[str]:
    """Get all regions for a list of countries. Countries can be a list of Country enums or a list of str."""
    all_regions = []
    if not countries:
        return all_regions

    if isinstance(countries[0], str):
        countries = [Country(country) for country in countries]

    for country in countries:
        if country == Country.belgium:
            all_regions.extend(RegionBe.get_all_values_str())
        elif country == Country.france:
            all_regions.extend(RegionFr.get_all_values_str())
        elif country == Country.netherlands:
            all_regions.extend(RegionNl.get_all_values_str())
        else:
            raise ValueError(f"Unknown country: `{country}`")

    return all_regions


def get_all_provinces(countries: list[Country] | list[str]) -> list[Province]:
    """Get all provinces for a list of countries. Countries can be a list of Country enums or a list of str."""
    all_provinces = []
    if not countries:
        return all_provinces

    if isinstance(countries[0], str):
        countries = [Country(country) for country in countries]

    if Country.belgium in countries:
        all_provinces.extend(ProvinceBe.get_all_values())
    if Country.france in countries:
        all_provinces.extend(ProvinceFr.get_all_values())
    if Country.netherlands in countries:
        all_provinces.extend(ProvinceNl.get_all_values())

    return all_provinces


def get_all_provinces_str(countries: list[Country] | list[str]) -> list[str]:
    """Get all provinces for a list of countries. Countries can be a list of Country enums or a list of str."""
    all_provinces = []
    if not countries:
        return all_provinces

    if isinstance(countries[0], str):
        countries = [Country(country) for country in countries]

    for country in countries:
        if country == Country.belgium:
            all_provinces.extend(ProvinceBe.get_all_values_str())
        elif country == Country.france:
            all_provinces.extend(ProvinceFr.get_all_values_str())
        elif country == Country.netherlands:
            all_provinces.extend(ProvinceNl.get_all_values_str())
        else:
            raise ValueError(f"Unknown country: `{country}`")

    return all_provinces


def coutries_to_enums(coutry_list: list[str]) -> list[Country]:
    """Format a list of countries as a string to a list of Country enums."""
    return [Country(country) for country in coutry_list]


def provinces_to_enums(province_list: list[str]) -> list[Province]:
    """Format a list of provinces as a string to a list of Province enums."""
    new_list = []
    for province in province_list:
        if province in ProvinceBe.get_all_values_str():
            province_enum = ProvinceBe(province)
        elif province in ProvinceFr.get_all_values_str():
            province_enum = ProvinceFr(province)
        elif province in ProvinceNl.get_all_values_str():
            province_enum = ProvinceNl(province)
        else:
            raise ValueError(f"Unknown province: `{province}`")
        new_list.append(province_enum)
    return new_list


def regions_to_enums(region_list: list[str]) -> list[Region]:
    """Format a list of regions as a string to a list of Region enums."""
    new_list = []
    for region in region_list:
        if region in RegionBe.get_all_values_str():
            region_enum = RegionBe(region)
        elif region in RegionFr.get_all_values_str():
            region_enum = RegionFr(region)
        elif region in RegionNl.get_all_values_str():
            region_enum = RegionNl(region)
        else:
            raise ValueError(f"Unknown region: `{region}`")
        new_list.append(region_enum)
    return new_list
